/**
 *  '$RCSfile: MetaCatServlet.java,v $'
 *    Purpose: A Class that implements a metadata catalog as a java Servlet
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Matt Jones, Dan Higgins, Jivka Bojilova, Chad Berkley
 *    Release: @release@
 *
 *   '$Author: jones $'
 *     '$Date: 2001/10/19 21:50:57 $'
 * '$Revision: 1.144 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import com.oreilly.servlet.multipart.FilePart;
import com.oreilly.servlet.multipart.MultipartParser;
import com.oreilly.servlet.multipart.ParamPart;
import com.oreilly.servlet.multipart.Part;

import java.io.File;
import java.io.PrintWriter;
import java.io.IOException;
import java.io.StringReader;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.ResourceBundle; 
import java.util.Random;
import java.util.PropertyResourceBundle;
import java.net.URL;
import java.net.MalformedURLException;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Connection;
import java.sql.SQLException;
import java.lang.reflect.*;
import java.net.*;
import java.util.zip.*;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletInputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpUtils;
import javax.servlet.ServletOutputStream;

import oracle.xml.parser.v2.XSLStylesheet;
import oracle.xml.parser.v2.XSLException;
import oracle.xml.parser.v2.XMLDocumentFragment;
import oracle.xml.parser.v2.XSLProcessor;

import org.xml.sax.SAXException;

/**
 * A metadata catalog server implemented as a Java Servlet
 *
 * <p>Valid parameters are:<br>
 * action=query -- query the values of all elements and attributes
 *                     and return a result set of nodes<br>
 * action=squery -- structured query (see pathquery.dtd)<br>
 * action=read -- read any metadata/data file from Metacat and from Internet<br>
 * action=insert -- insert an XML document into the database store<br>
 * action=update -- update an XML document that is in the database store<br>
 * action=delete --  delete an XML document from the database store<br>
 * action=validate -- vallidate the xml contained in valtext<br>
 * doctype -- document type list returned by the query (publicID)<br>
 * qformat=xml -- display resultset from query in XML<br>
 * qformat=html -- display resultset from query in HTML<br>
 * qformat=zip -- zip resultset from query<br>
 * docid=34 -- display the document with the document ID number 34<br>
 * doctext -- XML text of the document to load into the database<br>
 * acltext -- XML access text for a document to load into the database<br>
 * dtdtext -- XML DTD text for a new DTD to load into Metacat XML Catalog<br>
 * query -- actual query text (to go with 'action=query' or 'action=squery')<br>
 * valtext -- XML text to be validated<br>
 * abstractpath -- XPath in metadata document to read from<br>
 * action=getaccesscontrol -- retrieve acl info for Metacat document<br>
 * action=getdoctypes -- retrieve all doctypes (publicID)<br>
 * action=getdtdschema -- retrieve a DTD or Schema file<br>
 * action=getdataguide -- retrieve a Data Guide<br>
 * action=getprincipals -- retrieve a list of principals in XML<br>
 * datadoc -- data document name (id)<br>
 * <p>
 * The particular combination of parameters that are valid for each 
 * particular action value is quite specific.  This documentation
 * will be reorganized to reflect this information.
 */
public class MetaCatServlet extends HttpServlet {

  private ServletConfig config = null;
  private ServletContext context = null;
  private Hashtable connectionPool = new Hashtable();
  private String resultStyleURL = null;
  private String xmlcatalogfile = null;
  private String saxparser = null;
  private String datafilepath = null; 
  private File dataDirectory = null;
  private String servletpath = null; 
  private String htmlpath = null; 
  private PropertyResourceBundle options = null;
  private MetaCatUtil util = null;

  /**
   * Initialize the servlet by creating appropriate database connections
   */
  public void init( ServletConfig config ) throws ServletException {
    try {
      super.init( config );
      this.config = config;
      this.context = config.getServletContext(); 
      System.out.println("MetaCatServlet Initialize");

      util = new MetaCatUtil();

      // Get the configuration file information
      resultStyleURL = util.getOption("resultStyleURL");
      xmlcatalogfile = util.getOption("xmlcatalogfile");
      saxparser = util.getOption("saxparser");
      datafilepath = util.getOption("datafilepath");
      dataDirectory = new File(datafilepath);
      servletpath = util.getOption("servletpath");
      htmlpath = util.getOption("htmlpath");

// MOVED IT TO doGet() & doPost()
//      try {
//        // Open a pool of db connections
//        connectionPool = util.getConnectionPool();
//      } catch (Exception e) {
//        System.err.println("Error creating pool of database connections");
//        System.err.println(e.getMessage());
//      }
    } catch ( ServletException ex ) {
      throw ex;
    }
  }

  /**
   * Close all db connections from the pool
   */
  public void destroy() {
    
    if (util != null) {
        util.closeConnections();
    }
  }

  /** Handle "GET" method requests from HTTP clients */
  public void doGet (HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {

    // Process the data and send back the response
    handleGetOrPost(request, response);
  }

  /** Handle "POST" method requests from HTTP clients */
  public void doPost( HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {

    // Process the data and send back the response
    handleGetOrPost(request, response);
  }

  /**
   * Control servlet response depending on the action parameter specified
   */
  private void handleGetOrPost(HttpServletRequest request, 
                               HttpServletResponse response) 
                               throws ServletException, IOException 
  {

    if ( util == null ) {
        util = new MetaCatUtil(); 
    }
    if ( connectionPool.isEmpty() ) {
      try {
        // Open a pool of db connections
        connectionPool = util.getConnectionPool();
      } catch (Exception e) {
        System.err.println("Error creating pool of database connections in " +
                            " MetaCatServlet.handleGetOrPost");
        System.err.println(e.getMessage());
      }
    }    
    // Get a handle to the output stream back to the client
    //PrintWriter pwout = response.getWriter();
    //response.setContentType("text/html");

    // Deal with forms that are encoded as "multipart/form-data" differently
    // this is mainly used for uploading non-xml files using that may be binary
    String ctype = request.getContentType();
    if (ctype != null && ctype.startsWith("multipart/form-data")) {
      handleMultipartForm(request, response);
    } else {
      // This probably means the data is "application/x-www-url-encoded", we hope :-)
  
      String name = null;
      String[] value = null;
      String[] docid = new String[3];
      Hashtable params = new Hashtable();
      Enumeration paramlist = request.getParameterNames();
      while (paramlist.hasMoreElements()) {
        name = (String)paramlist.nextElement();
        value = request.getParameterValues(name);
  
        // Decode the docid and mouse click information
        if (name.endsWith(".y")) {
          docid[0] = name.substring(0,name.length()-2);
          params.put("docid", docid);
          name = "ypos";
        }
        if (name.endsWith(".x")) {
          name = "xpos";
        } 
  
        params.put(name,value); 
      }  
      
      //if the user clicked on the input images, decode which image
      //was clicked then set the action.
      String action = ((String[])params.get("action"))[0];  
      util.debugMessage("Line 213: Action is: " + action);
  
      // This block handles session management for the servlet
      // by looking up the current session information for all actions
      // other than "login" and "logout"
      String username = null;
      String password = null;
      String[] groupnames = null;
      String sess_id = null;
  
      // handle login action
      if (action.equals("login")) {
  
        handleLoginAction(response.getWriter(), params, request, response);
  
      // handle logout action  
      } else if (action.equals("logout")) {
  
        handleLogoutAction(response.getWriter(), params, request, response);
  
      // aware of session expiration on every request  
      } else {   
  
        HttpSession sess = request.getSession(true);
        if (sess.isNew()) { 
          // session expired or has not been stored b/w user requests
          username = "public";
          sess.setAttribute("username", username);
        } else {
          username = (String)sess.getAttribute("username");
          password = (String)sess.getAttribute("password");
          groupnames = (String[])sess.getAttribute("groupnames");
          try {
            sess_id = (String)sess.getId();
          } catch(IllegalStateException ise) {
            System.out.println("error in handleGetOrPost: this shouldn't " +
                               "happen: the session should be valid: " + 
                               ise.getMessage());
          }
        }  
      }    
  
      // Now that we know the session is valid, we can delegate the request
      // to a particular action handler
      if(action.equals("query")) {
        handleQuery(response.getWriter(),params,response,username,groupnames);
      } else if(action.equals("squery")) {
        if(params.containsKey("query")) {
          handleSQuery(response.getWriter(),params,response,username,groupnames); 
        } else {
          PrintWriter out = response.getWriter();
          out.println("Illegal action squery without \"query\" parameter");
        }
      } else if (action.equals("read")) {
        handleReadAction(params, response, username, groupnames);
      } else if (action.equals("insert") || action.equals("update")) {
        PrintWriter out = response.getWriter();
        if ( (username != null) &&  !username.equals("public") ) {
          handleInsertOrUpdateAction(out,params,response,username,groupnames);
        } else {  
          out.println("Permission denied for " + action);
        }  
      } else if (action.equals("delete")) {
        PrintWriter out = response.getWriter();
        if ( (username != null) &&  !username.equals("public") ) {
          handleDeleteAction(out, params, response, username, groupnames);
        } else {  
          out.println("Permission denied for " + action);
        }  
      } else if (action.equals("validate")) {
        PrintWriter out = response.getWriter();
        handleValidateAction(out, params, response); 
      } else if (action.equals("getaccesscontrol")) {
        PrintWriter out = response.getWriter();
        handleGetAccessControlAction(out,params,response,username,groupnames);
      } else if (action.equals("getprincipals")) {
        PrintWriter out = response.getWriter();
        handleGetPrincipalsAction(out, username, password);  
      } else if (action.equals("getdoctypes")) {
        PrintWriter out = response.getWriter();
        handleGetDoctypesAction(out, params, response);  
      } else if (action.equals("getdtdschema")) {
        PrintWriter out = response.getWriter();
        handleGetDTDSchemaAction(out, params, response);  
      } else if (action.equals("getdataguide")) {
        PrintWriter out = response.getWriter();
        handleGetDataGuideAction(out, params, response);  
      } else if (action.equals("getlastdocid")) {
        PrintWriter out = response.getWriter();
        handleGetMaxDocidAction(out, params, response);  
      } else if (action.equals("login") || action.equals("logout")) {
      } else if (action.equals("protocoltest")) {
        String testURL = "metacat://dev.nceas.ucsb.edu/NCEAS.897766.9";
        try {
          testURL = ((String[])params.get("url"))[0];
        } catch (Throwable t) {
        }
        String phandler = System.getProperty("java.protocol.handler.pkgs");
        response.setContentType("text/html");
        PrintWriter out = response.getWriter();
        out.println("<body bgcolor=\"white\">");
        out.println("<p>Handler property: <code>" + phandler + "</code></p>");
        out.println("<p>Starting test for:<br>");
        out.println("    " + testURL + "</p>");
        try {
          URL u = new URL(testURL);
          out.println("<pre>");
          out.println("Protocol: " + u.getProtocol());
          out.println("    Host: " + u.getHost());
          out.println("    Port: " + u.getPort());
          out.println("    Path: " + u.getPath());
          out.println("     Ref: " + u.getRef());
          String pquery = u.getQuery();
          out.println("   Query: " + pquery);
          out.println("  Params: ");
          if (pquery != null) {
            Hashtable qparams = util.parseQuery(u.getQuery());
            for (Enumeration en = qparams.keys(); en.hasMoreElements(); ) {
              String pname = (String)en.nextElement();
              String pvalue = (String)qparams.get(pname);
              out.println("    " + pname + ": " + pvalue);
            }
          }
          out.println("</pre>");
          out.println("</body>");
          out.close();
        } catch (MalformedURLException mue) {
          System.out.println("bad url from MetacatServlet.handleGetOrPost");
          out.println(mue.getMessage());
          mue.printStackTrace(out);
          out.close();
        }
      } else {
        PrintWriter out = response.getWriter();
        out.println("<?xml version=\"1.0\"?>");
        out.println("<error>");
        out.println("Error: action not registered.  Please report this error.");
        out.println("</error>");
      }
      
      util.closeConnections();
      // Close the stream to the client
      // out.close();
    }
  }
  
  // LOGIN & LOGOUT SECTION
  /** 
   * Handle the login request. Create a new session object.
   * Do user authentication through the session.
   */
  private void handleLoginAction(PrintWriter out, Hashtable params, 
               HttpServletRequest request, HttpServletResponse response) {

    AuthSession sess = null;
    String un = ((String[])params.get("username"))[0];
    String pw = ((String[])params.get("password"))[0];
    String action = ((String[])params.get("action"))[0];
    String qformat = ((String[])params.get("qformat"))[0];
    
    try {
      sess = new AuthSession();
    } catch (Exception e) {
      System.out.println("error in MetacatServlet.handleLoginAction: " +
                          e.getMessage());
      out.println(e.getMessage());
      return;
    }
    
    boolean isValid = sess.authenticate(request, un, pw);

    // format and transform the output
    if (qformat.equals("xml")) {
      response.setContentType("text/xml");
      out.println(sess.getMessage()); 
    } else {
      Connection conn = null;
      try {
        conn = util.getConnection();
        DBTransform trans = new DBTransform(conn);
        response.setContentType("text/html");
        trans.transformXMLDocument(sess.getMessage(), "-//NCEAS//login//EN",
                                   "-//W3C//HTML//EN", qformat, out);
        util.returnConnection(conn); 
      } catch(Exception e) {
        util.returnConnection(conn); 
      } 
      
    // any output is returned  
    }
  }    

  /** 
   * Handle the logout request. Close the connection.
   */
  private void handleLogoutAction(PrintWriter out, Hashtable params, 
               HttpServletRequest request, HttpServletResponse response) {

    String qformat = ((String[])params.get("qformat"))[0];

    // close the connection
    HttpSession sess = request.getSession(false);
    if (sess != null) { sess.invalidate();  }    

    // produce output
    StringBuffer output = new StringBuffer();
    output.append("<?xml version=\"1.0\"?>");
    output.append("<logout>");
    output.append("User logged out");
    output.append("</logout>");

    //format and transform the output
    if (qformat.equals("xml")) {
      response.setContentType("text/xml");
      out.println(output.toString()); 
    } else {
      Connection conn = null;
      try {
        conn = util.getConnection();
        DBTransform trans = new DBTransform(conn);
        response.setContentType("text/html");
        trans.transformXMLDocument(output.toString(), "-//NCEAS//login//EN", 
                                   "-//W3C//HTML//EN", qformat, out);
        util.returnConnection(conn); 
      } catch(Exception e) {
        util.returnConnection(conn); 
      } 
    }
  }
  // END OF LOGIN & LOGOUT SECTION
  
  // SQUERY & QUERY SECTION
  /**      
   * Retreive the squery xml, execute it and display it
   *
   * @param out the output stream to the client
   * @param params the Hashtable of parameters that should be included
   * in the squery.
   * @param response the response object linked to the client
   * @param conn the database connection 
   */
  protected void handleSQuery(PrintWriter out, Hashtable params, 
                 HttpServletResponse response, String user, String[] groups)
  { 
    String xmlquery = ((String[])params.get("query"))[0];
    String qformat = ((String[])params.get("qformat"))[0];
    String resultdoc = null;
    
    Hashtable doclist = runQuery(xmlquery, user, groups);

    resultdoc = createResultDocument(doclist, transformQuery(xmlquery));
    
    //format and transform the results                                        
    if(qformat.equals("xml")) {
      response.setContentType("text/xml");
      out.println(resultdoc);
    } else {
      transformResultset(resultdoc, response, out, qformat);
    }
  }
  
   /**
    * Create the xml query, execute it and display the results.
    *
    * @param out the output stream to the client
    * @param params the Hashtable of parameters that should be included
    * in the squery.
    * @param response the response object linked to the client
    */ 
  protected void handleQuery(PrintWriter out, Hashtable params, 
                 HttpServletResponse response, String user, String[] groups)
  {
    //create the query and run it
    String xmlquery = DBQuery.createSQuery(params);
    Hashtable doclist = runQuery(xmlquery, user, groups);
    String qformat = ((String[])params.get("qformat"))[0];
    String resultdoc = null;
    
    resultdoc = createResultDocument(doclist, transformQuery(params));

    //format and transform the results                                        
    if(qformat.equals("xml")) {
      response.setContentType("text/xml");
      out.println(resultdoc);
    } else { 
      transformResultset(resultdoc, response, out, qformat);
    }
  }
  
  /**
   * Removes the <?xml version="x"?> tag from the beginning of xmlquery
   * so it can properly be placed in the <query> tag of the resultset.
   * This method is overwritable so that other applications can customize
   * the structure of what is in the <query> tag.
   * 
   * @param xmlquery is the query to remove the <?xml version="x"?> tag from.
   */
  protected String transformQuery(Hashtable params)
  {
    //DBQuery.createSQuery is a re-calling of a previously called 
    //function but it is necessary
    //so that overriding methods have access to the params hashtable
    String xmlquery = DBQuery.createSQuery(params);
    //the <?xml version="1.0"?> tag is the first 22 characters of the
    xmlquery = xmlquery.trim();
    int index = xmlquery.indexOf("?>");
    return xmlquery.substring(index + 2, xmlquery.length());
  }
  
  /**
   * removes the <?xml version="1.0"?> tag from the beginning.  This takes a
   * string as a param instead of a hashtable.
   * 
   * @param xmlquery a string representing a query.
   */
  protected String transformQuery(String xmlquery)
  {
    xmlquery = xmlquery.trim();
    int index = xmlquery.indexOf("?>");
    return xmlquery.substring(index + 2, xmlquery.length());
  }
  
  /**
   * Run the query and return a hashtable of results.
   *
   * @param xmlquery the query to run
   */
  private Hashtable runQuery(String xmlquery, String user, String[] groups)
  {
    Hashtable doclist=null;
    Connection conn = null;
    try
    {
      conn = util.getConnection();
      DBQuery queryobj = new DBQuery(conn, saxparser);
      doclist = queryobj.findDocuments(new StringReader(xmlquery),user,groups);
      util.returnConnection(conn);
      return doclist;
    } 
    catch (Exception e) 
    {
      util.returnConnection(conn); 
      util.debugMessage("Error in MetacatServlet.runQuery: " + e.getMessage());
      doclist = null;
      return doclist;
    }    
  }
  
  /**
   * Transorms an xml resultset document to html and sends it to the browser
   *
   * @param resultdoc the string representation of the document that needs
   * to be transformed.
   * @param response the HttpServletResponse object bound to the client.
   * @param out the output stream to the client
   * @param qformat the name of the style-set to use for transformations
   */ 
  protected void transformResultset(String resultdoc, 
                                    HttpServletResponse response,
                                    PrintWriter out, String qformat)
  {
    Connection conn = null;
    try {
      conn = util.getConnection();
      DBTransform trans = new DBTransform(conn);
      response.setContentType("text/html");
      trans.transformXMLDocument(resultdoc, "-//NCEAS//resultset//EN", 
                                 "-//W3C//HTML//EN", qformat, out);
      util.returnConnection(conn); 
    }
    catch(Exception e)
    {
      util.returnConnection(conn); 
    } 
  }
  
  /**
   * Transforms a hashtable of documents to an xml or html result.
   *
   * @param doclist- the hashtable to transform
   * @param xmlquery- the query that returned the doclist result
   */
  protected String createResultDocument(Hashtable doclist, String xmlquery)
  {
    // Create a buffer to hold the xml result
    StringBuffer resultset = new StringBuffer();
 
    // Print the resulting root nodes 
    String docid = null;
    String document = null;
    resultset.append("<?xml version=\"1.0\"?>\n");
    resultset.append("<resultset>\n");
      
    resultset.append("  <query>" + xmlquery + "</query>");   

    if(doclist != null)
    {
      Enumeration doclistkeys = doclist.keys(); 
      while (doclistkeys.hasMoreElements()) 
      {
        docid = (String)doclistkeys.nextElement();
        document = (String)doclist.get(docid);
        resultset.append("  <document>" + document + "</document>");
      }
    }

    resultset.append("</resultset>");
    return resultset.toString();
  }
  // END OF SQUERY & QUERY SECTION
  
  // READ SECTION
  /** 
   * Handle the "read" request of metadata/data files from Metacat
   * or any files from Internet;
   * transformed metadata XML document into HTML presentation if requested;
   * zip files when more than one were requested.
   *
   * @param params the Hashtable of HTTP request parameters
   * @param response the HTTP response object linked to the client
   * @param user the username sent the request
   * @param groups the user's groupnames
   */
  private void handleReadAction(Hashtable params, HttpServletResponse response,
                                String user, String[] groups) 
  {
    ServletOutputStream out = null;
    ZipOutputStream zout = null;
    
    try {
      String[] docs = new String[0];
      String docid = "";
      String qformat = "";
      String abstrpath = null;
      boolean zip = false;
      // read the params
      if (params.containsKey("docid")) {
        docs = (String[])params.get("docid");
      }
      if (params.containsKey("qformat")) {
        qformat = ((String[])params.get("qformat"))[0];
      }
      if (params.containsKey("abstractpath")) {
        abstrpath = ((String[])params.get("abstractpath"))[0];
        if ( !abstrpath.equals("") && (abstrpath != null) ) {
          viewAbstract(response, abstrpath, docs[0]);
          return;
        }
      }
      if ( (docs.length > 1) || qformat.equals("zip") ) {
        zip = true;
        out = response.getOutputStream();
        response.setContentType("application/zip"); //MIME type
        zout = new ZipOutputStream(out);
      }
      // go through the list of docs to read
      for (int i=0; i < docs.length; i++ ) {
        try {

          URL murl = new URL(docs[i]);
          Hashtable murlQueryStr = util.parseQuery(murl.getQuery());
          // case docid="http://.../?docid=aaa" 
          // or docid="metacat://.../?docid=bbb"
          if (murlQueryStr.containsKey("docid")) {
            // get only docid, eliminate the rest
            docid = (String)murlQueryStr.get("docid");
            if ( zip ) {
              addDocToZip(docid, zout);
            } else {
              readFromMetacat(response, docid, qformat, abstrpath,
                              user, groups, zip, zout);
            }

          // case docid="http://.../filename"
          } else {
            docid = docs[i];
            if ( zip ) {
              addDocToZip(docid, zout);
            } else {
              readFromURLConnection(response, docid);
            }
          }

        // case docid="ccc"
        } catch (MalformedURLException mue) {
          docid = docs[i];
          if ( zip ) {
            addDocToZip(docid, zout);
          } else {
            readFromMetacat(response, docid, qformat, abstrpath,
                            user, groups, zip, zout);
          }
        }
        
      } /* end for */
      
      if ( zip ) {
        zout.finish(); //terminate the zip file
        zout.close();  //close the zip stream
      }
      
        /*
    } catch (ClassNotFoundException cnfe) {
    } catch (IOException ioe1) {
    } catch (SQLException se) {
    } catch (McdbException mcdbe) {
    } catch (Exception e) {
      */
    } catch (Exception e) {
      try {
        response.setContentType("text/xml"); //MIME type
        //PrintWriter pw = response.getWriter();
        if (out != null) {
            PrintWriter pw = new PrintWriter(out);
            pw.println("<?xml version=\"1.0\"?>");
            pw.println("<error>");
            pw.println(e.getMessage());
            pw.println("</error>");
            pw.close();
        }
        //if ( out != null ) { out.close(); }
        if ( zout != null ) { zout.close(); }
      } catch (IOException ioe) {
        System.out.println("Problem with the servlet output " +
                           "in MetacatServlet.handleReadAction: " +
                           ioe.getMessage());
        ioe.printStackTrace(System.out);
        
      }

      System.out.println("Error in MetacatServlet.handleReadAction: " +
                         e.getMessage());
      e.printStackTrace(System.out);
    }
    
  }
  
  // read metadata or data from Metacat
  private void readFromMetacat(HttpServletResponse response, String docid,
                               String qformat, String abstrpath, String user,
                               String[] groups, boolean zip, ZipOutputStream zout)
               throws ClassNotFoundException, IOException, SQLException, 
                      McdbException, Exception
  {
    Connection conn = null;
    try {
      conn = util.getConnection();
      DocumentImpl doc = new DocumentImpl(conn, docid);
      
      if ( doc.getRootNodeID() == 0 ) {
        // this is data file
        ServletOutputStream out = response.getOutputStream(); 
        String filepath = util.getOption("datafilepath");
        if(!filepath.endsWith("/")) {
          filepath += "/";
        }
        String filename = filepath + docid;      //MIME type
        String contentType = getServletContext().getMimeType(filename);
        if (contentType == null) {
          if (filename.endsWith(".xml")) {
            contentType="text/xml";
          } else if (filename.endsWith(".css")) {
            contentType="text/css";
          } else if (filename.endsWith(".dtd")) {
            contentType="text/plain";
          } else if (filename.endsWith(".xsd")) {
            contentType="text/xml";
          } else if (filename.endsWith("/")) {
            contentType="text/html";
          } else {
            File f = new File(filename);
            if ( f.isDirectory() ) {
              contentType="text/html";
            } else {
              contentType="application/octet-stream";
            }
          }
        }
        response.setContentType(contentType);
        // if we decide to use "application/octet-stream" for all data returns
        // response.setContentType("application/octet-stream");
        FileInputStream fin = null;
        try {
          fin = new FileInputStream(filename);
          byte[] buf = new byte[4 * 1024]; // 4K buffer
          int b = fin.read(buf);
          while (b != -1) {
            out.write(buf, 0, b);
            b = fin.read(buf);
          }
        } finally {
          if (fin != null) fin.close();
        }

      } else {
        // this is metadata doc
        if ( qformat.equals("xml") ) { 
          // set content type first
          response.setContentType("text/xml");   //MIME type
          PrintWriter out = response.getWriter();
          doc.toXml(out);
        } else {
          response.setContentType("text/html");  //MIME type
          PrintWriter out = response.getWriter();
    
          // Look up the document type
          String doctype = doc.getDoctype();
          // Transform the document to the new doctype
          DBTransform dbt = new DBTransform(conn);
          dbt.transformXMLDocument(doc.toString(),
                                   doctype,"-//W3C//HTML//EN", qformat, out);
        }
      
      }
    } finally {
      util.returnConnection(conn);
    }
    
  }
  
  // read data from URLConnection
  private void readFromURLConnection(HttpServletResponse response, String docid)
               throws IOException, MalformedURLException
  {
    ServletOutputStream out = response.getOutputStream(); 
    String contentType = getServletContext().getMimeType(docid); //MIME type
    if (contentType == null) {
      if (docid.endsWith(".xml")) {
        contentType="text/xml";
      } else if (docid.endsWith(".css")) {
        contentType="text/css";
      } else if (docid.endsWith(".dtd")) {
        contentType="text/plain";
      } else if (docid.endsWith(".xsd")) {
        contentType="text/xml";
      } else if (docid.endsWith("/")) {
        contentType="text/html";
      } else {
        File f = new File(docid);
        if ( f.isDirectory() ) {
          contentType="text/html";
        } else {
          contentType="application/octet-stream";
        }
      }
    }
    response.setContentType(contentType);
    // if we decide to use "application/octet-stream" for all data returns
    // response.setContentType("application/octet-stream");

    // this is http url
    URL url = new URL(docid);
    BufferedInputStream bis = null;
    try {
      bis = new BufferedInputStream(url.openStream());
      byte[] buf = new byte[4 * 1024]; // 4K buffer
      int b = bis.read(buf);
      while (b != -1) {
        out.write(buf, 0, b);
        b = bis.read(buf);
      }
    } finally {
      if (bis != null) bis.close();
    }
    
  }
  
  // read file/doc and write to ZipOutputStream
  private void addDocToZip(String docid, ZipOutputStream zout)
               throws ClassNotFoundException, IOException, SQLException, 
                      McdbException, Exception
  {
    byte[] bytestring = null;
    ZipEntry zentry = null;

    try {
      URL url = new URL(docid);

      // this http url; read from URLConnection; add to zip
      zentry = new ZipEntry(docid);
      zout.putNextEntry(zentry);
      BufferedInputStream bis = null;
      try {
        bis = new BufferedInputStream(url.openStream());
        byte[] buf = new byte[4 * 1024]; // 4K buffer
        int b = bis.read(buf);
        while(b != -1) {
          zout.write(buf, 0, b);
          b = bis.read(buf);
        }
      } finally {
        if (bis != null) bis.close();
      }
      zout.closeEntry();

    } catch (MalformedURLException mue) {
      
      // this is metacat doc (data file or metadata doc)
      Connection conn = null;
      try {
        conn = util.getConnection();
        DocumentImpl doc = new DocumentImpl(conn, docid);
      
        if ( doc.getRootNodeID() == 0 ) {
          // this is data file; add file to zip
          String filepath = util.getOption("datafilepath");
          if(!filepath.endsWith("/")) {
            filepath += "/";
          }
          String filename = filepath + doc.getDocname();
          zentry = new ZipEntry(filename);
          zout.putNextEntry(zentry);
          FileInputStream fin = null;
          try {
            fin = new FileInputStream(filename);
            byte[] buf = new byte[4 * 1024]; // 4K buffer
            int b = fin.read(buf);
            while (b != -1) {
              zout.write(buf, 0, b);
              b = fin.read(buf);
            }
          } finally {
            if (fin != null) fin.close();
          }
          zout.closeEntry();

        } else {
          // this is metadata doc; add doc to zip
          bytestring = doc.toString().getBytes();
          zentry = new ZipEntry(docid + ".xml");
          zentry.setSize(bytestring.length);
          zout.putNextEntry(zentry);
          zout.write(bytestring, 0, bytestring.length);
          zout.closeEntry();
        }
      } finally {
        util.returnConnection(conn);
      }
      
    }
      
  }
  
  // view abstract within document
  private void viewAbstract(HttpServletResponse response,
                            String abstractpath, String docid)
               throws ClassNotFoundException, IOException, SQLException,
                      McdbException, Exception
  {
    Connection conn = null;
    try {
      conn = util.getConnection();
    
      Object[] abstracts = DBQuery.getNodeContent(abstractpath, docid, conn);
    
      response.setContentType("text/html");  //MIME type
      PrintWriter out = response.getWriter();
      out.println("<html><head><title>Abstract</title></head>");
      out.println("<body bgcolor=\"white\"><h1>Abstract</h1>");
      for (int i=0; i<abstracts.length; i++) {
        out.println("<p>" + (String)abstracts[i] + "</p>");
      }
      out.println("</body></html>");

    } finally {
      util.returnConnection(conn);
    }
  }
  // END OF READ SECTION
    
  // INSERT/UPDATE SECTION
  /** 
   * Handle the database putdocument request and write an XML document 
   * to the database connection
   */
  private void handleInsertOrUpdateAction(PrintWriter out, Hashtable params, 
               HttpServletResponse response, String user, String[] groups) {

    Connection conn = null;

    try {
      // Get the document indicated
      String[] doctext = (String[])params.get("doctext");

      String pub = null;
      if (params.containsKey("public")) {
        pub = ((String[])params.get("public"))[0];
      }

      StringReader dtd = null;
      if (params.containsKey("dtdtext")) {
        String[] dtdtext = (String[])params.get("dtdtext");
        try {
          if ( !dtdtext[0].equals("") ) {
            dtd = new StringReader(dtdtext[0]);
          }
        } catch (NullPointerException npe) {}
      }
      
      StringReader xml = null;
      boolean validate = false;
      try {
        // look inside XML Document for <!DOCTYPE ... PUBLIC/SYSTEM ... > 
        // in order to decide whether to use validation parser
        validate = validateXML(doctext[0]);
        xml = new StringReader(doctext[0]);

        String[] action = (String[])params.get("action");
        String[] docid = (String[])params.get("docid");
        String newdocid = null;

        String doAction = null;
        if (action[0].equals("insert")) {
          doAction = "INSERT";
        } else if (action[0].equals("update")) {
          doAction = "UPDATE";
        }

        try {
          // get a connection from the pool
          conn = util.getConnection();

          // write the document to the database
          try {
            String accNumber = docid[0];
            if (accNumber.equals("")) {
              accNumber = null;
            }
            newdocid = DocumentImpl.write(conn, xml, pub, dtd, doAction,
                                          accNumber, user, groups, validate);
          } catch (NullPointerException npe) {
            newdocid = DocumentImpl.write(conn, xml, pub, dtd, doAction,
                                          null, user, groups, validate);
          }
        } finally {
          util.returnConnection(conn);
        }    

        // set content type and other response header fields first
        response.setContentType("text/xml");
        out.println("<?xml version=\"1.0\"?>");
        out.println("<success>");
        out.println("<docid>" + newdocid + "</docid>"); 
        out.println("</success>");

      } catch (NullPointerException npe) {
        response.setContentType("text/xml");
        out.println("<?xml version=\"1.0\"?>");
        out.println("<error>");
        out.println(npe.getMessage()); 
        out.println("</error>");
      }
    } catch (Exception e) {
      response.setContentType("text/xml");
      out.println("<?xml version=\"1.0\"?>");
      out.println("<error>");
      out.println(e.getMessage()); 
      if (e instanceof SAXException) {
        Exception e2 = ((SAXException)e).getException();
        out.println("<error>");
        try
        {
          out.println(e2.getMessage());
        }
        catch(NullPointerException npe)
        {
          out.println(e.getMessage());
        }
        out.println("</error>");
      }
      //e.printStackTrace(out);
      out.println("</error>");
    }
  }

  /** 
   * Parse XML Document to look for <!DOCTYPE ... PUBLIC/SYSTEM ... > 
   * in order to decide whether to use validation parser
   */
  private static boolean validateXML(String xmltext) throws IOException {
    
    StringReader xmlreader = new StringReader(xmltext);
    StringBuffer cbuff = new StringBuffer();
    java.util.Stack st = new java.util.Stack();
    boolean validate = false;
    int c;
    int inx;
    
    // read from the stream until find the keywords
    while ( (st.empty() || st.size()<4) && ((c = xmlreader.read()) != -1) ) {
      cbuff.append((char)c);

      // "<!DOCTYPE" keyword is found; put it in the stack
      if ( (inx = cbuff.toString().indexOf("<!DOCTYPE")) != -1 ) {
        cbuff = new StringBuffer();
        st.push("<!DOCTYPE");
      }
      // "PUBLIC" keyword is found; put it in the stack
      if ( (inx = cbuff.toString().indexOf("PUBLIC")) != -1 ) {
        cbuff = new StringBuffer();
        st.push("PUBLIC");
      }
      // "SYSTEM" keyword is found; put it in the stack
      if ( (inx = cbuff.toString().indexOf("SYSTEM")) != -1 ) {
        cbuff = new StringBuffer();
        st.push("SYSTEM");
      }
      // ">" character is found; put it in the stack
      // ">" is found twice: fisrt from <?xml ...?> 
      // and second from <!DOCTYPE ... >
      if ( (inx = cbuff.toString().indexOf(">")) != -1 ) {
        cbuff = new StringBuffer();
        st.push(">");
      }
    }

    // close the stream
    xmlreader.close();

    // check the stack whether it contains the keywords:
    // "<!DOCTYPE", "PUBLIC" or "SYSTEM", and ">" in this order
    if ( st.size() == 4 ) {
      if ( ((String)st.pop()).equals(">") &&
           ( ((String)st.peek()).equals("PUBLIC") |
             ((String)st.pop()).equals("SYSTEM") ) &&
           ((String)st.pop()).equals("<!DOCTYPE") )  {
        validate = true;
      }
    }

System.out.println("Validation is " + validate);
    return validate;
  }
  // END OF INSERT/UPDATE SECTION

  // DELETE SECTION
  /** 
   * Handle the database delete request and delete an XML document 
   * from the database connection
   */
  private void handleDeleteAction(PrintWriter out, Hashtable params, 
               HttpServletResponse response, String user, String[] groups) {

    String[] docid = (String[])params.get("docid");
    Connection conn = null;

    // delete the document from the database
    try {
      // get a connection from the pool
      conn = util.getConnection();
                                      // NOTE -- NEED TO TEST HERE
                                      // FOR EXISTENCE OF DOCID PARAM
                                      // BEFORE ACCESSING ARRAY
      try { 
        DocumentImpl.delete(conn, docid[0], user, groups);
        response.setContentType("text/xml");
        out.println("<?xml version=\"1.0\"?>");
        out.println("<success>");
        out.println("Document deleted."); 
        out.println("</success>");
      } catch (AccessionNumberException ane) {
        response.setContentType("text/xml");
        out.println("<?xml version=\"1.0\"?>");
        out.println("<error>");
        out.println("Error deleting document!!!");
        out.println(ane.getMessage()); 
        out.println("</error>");
      }
    } catch (Exception e) {
      response.setContentType("text/xml");
      out.println("<?xml version=\"1.0\"?>");
      out.println("<error>");
      out.println(e.getMessage()); 
      out.println("</error>");
    } finally {
      util.returnConnection(conn);
    }  
  }
  // END OF DELETE SECTION
  
  // VALIDATE SECTION
  /** 
   * Handle the validation request and return the results to the requestor
   */
  private void handleValidateAction(PrintWriter out, Hashtable params, 
               HttpServletResponse response) {

    // Get the document indicated
    String valtext = null;
    
    try {
      valtext = ((String[])params.get("valtext"))[0];
    } catch (Exception nullpe) {

      Connection conn = null;
      String docid = null;
      try {
        // Find the document id number
        docid = ((String[])params.get("docid"))[0]; 

        // get a connection from the pool
        conn = util.getConnection();

        // Get the document indicated from the db
        DocumentImpl xmldoc = new DocumentImpl(conn, docid);
        valtext = xmldoc.toString();

      } catch (NullPointerException npe) {
        response.setContentType("text/xml");
        out.println("<error>Error getting document ID: " + docid + "</error>");
        if ( conn != null ) { util.returnConnection(conn); }
        return;
      } catch (Exception e) {
        response.setContentType("text/html");
        out.println(e.getMessage()); 
      } finally {
        util.returnConnection(conn);
      }  
    }

    Connection conn = null;
    try {
      // get a connection from the pool
      conn = util.getConnection();
      DBValidate valobj = new DBValidate(saxparser,conn);
      boolean valid = valobj.validateString(valtext);

      // set content type and other response header fields first
      response.setContentType("text/xml");
      out.println(valobj.returnErrors());

    } catch (NullPointerException npe2) {
      // set content type and other response header fields first
      response.setContentType("text/xml");
      out.println("<error>Error validating document.</error>"); 
    } catch (Exception e) {
      response.setContentType("text/html");
      out.println(e.getMessage()); 
    } finally {
      util.returnConnection(conn);
    }  
  }
  // END OF VALIDATE SECTION
 
  // OTHER ACTION HANDLERS
  
  /** 
   * Handle "getaccesscontrol" action.
   * Read Access Control List from db connection in XML format
   */
  private void handleGetAccessControlAction(PrintWriter out, Hashtable params, 
                                       HttpServletResponse response, 
                                       String username, String[] groupnames) {

    Connection conn = null;
    String docid = ((String[])params.get("docid"))[0];
    
    try {

        // get connection from the pool
        conn = util.getConnection();
        AccessControlList aclobj = new AccessControlList(conn);
        String acltext = aclobj.getACL(docid, username, groupnames);
        out.println(acltext);

    } catch (Exception e) {
      out.println("<?xml version=\"1.0\"?>");
      out.println("<error>");
      out.println(e.getMessage());
      out.println("</error>");
    } finally {
      util.returnConnection(conn);
    }  
    
  }

  /** 
   * Handle the "getprincipals" action.
   * Read all principals from authentication scheme in XML format
   */
  private void handleGetPrincipalsAction(PrintWriter out, String user,
                                         String password) {

    Connection conn = null;

    try {

        // get connection from the pool
        AuthSession auth = new AuthSession();
        String principals = auth.getPrincipals(user, password);
        out.println(principals);

    } catch (Exception e) {
      out.println("<?xml version=\"1.0\"?>");
      out.println("<error>");
      out.println(e.getMessage());
      out.println("</error>");
    } finally {
      util.returnConnection(conn);
    }  
    
  }

  /** 
   * Handle "getdoctypes" action.
   * Read all doctypes from db connection in XML format
   */
  private void handleGetDoctypesAction(PrintWriter out, Hashtable params, 
                                       HttpServletResponse response) {

    Connection conn = null;
    
    try {

        // get connection from the pool
        conn = util.getConnection();
        DBUtil dbutil = new DBUtil(conn);
        String doctypes = dbutil.readDoctypes();
        out.println(doctypes);

    } catch (Exception e) {
      out.println("<?xml version=\"1.0\"?>");
      out.println("<error>");
      out.println(e.getMessage());
      out.println("</error>");
    } finally {
      util.returnConnection(conn);
    }  
    
  }

  /** 
   * Handle the "getdtdschema" action.
   * Read DTD or Schema file for a given doctype from Metacat catalog system
   */
  private void handleGetDTDSchemaAction(PrintWriter out, Hashtable params,
                                        HttpServletResponse response) {

    Connection conn = null;
    String doctype = null;
    String[] doctypeArr = (String[])params.get("doctype");

    // get only the first doctype specified in the list of doctypes
    // it could be done for all doctypes in that list
    if (doctypeArr != null) {
        doctype = ((String[])params.get("doctype"))[0]; 
    }

    try {

        // get connection from the pool
        conn = util.getConnection();
        DBUtil dbutil = new DBUtil(conn);
        String dtdschema = dbutil.readDTDSchema(doctype);
        out.println(dtdschema);

    } catch (Exception e) {
      out.println("<?xml version=\"1.0\"?>");
      out.println("<error>");
      out.println(e.getMessage());
      out.println("</error>");
    } finally {
      util.returnConnection(conn);
    }  
    
  }

  /** 
   * Handle the "getdataguide" action.
   * Read Data Guide for a given doctype from db connection in XML format
   */
  private void handleGetDataGuideAction(PrintWriter out, Hashtable params, 
                                        HttpServletResponse response) {

    Connection conn = null;
    String doctype = null;
    String[] doctypeArr = (String[])params.get("doctype");

    // get only the first doctype specified in the list of doctypes
    // it could be done for all doctypes in that list
    if (doctypeArr != null) {
        doctype = ((String[])params.get("doctype"))[0]; 
    }

    try {

        // get connection from the pool
        conn = util.getConnection();
        DBUtil dbutil = new DBUtil(conn);
        String dataguide = dbutil.readDataGuide(doctype);
        out.println(dataguide);

    } catch (Exception e) {
      out.println("<?xml version=\"1.0\"?>");
      out.println("<error>");
      out.println(e.getMessage());
      out.println("</error>");
    } finally {
      util.returnConnection(conn);
    }  
    
  }

  /** 
   * Handle the "getlastdocid" action.
   * Get the latest docid with rev number from db connection in XML format
   */
  private void handleGetMaxDocidAction(PrintWriter out, Hashtable params, 
                                        HttpServletResponse response) {

    Connection conn = null;
    String scope = ((String[])params.get("scope"))[0];
    if (scope == null) {
        scope = ((String[])params.get("username"))[0];
    }

    try {

        // get connection from the pool
        conn = util.getConnection();
        DBUtil dbutil = new DBUtil(conn);
        String lastDocid = dbutil.getMaxDocid(scope);
        out.println("<?xml version=\"1.0\"?>");
        out.println("<lastDocid>");
        out.println("  <scope>" + scope + "</scope>");
        out.println("  <docid>" + lastDocid + "</docid>");
        out.println("</lastDocid>");

    } catch (Exception e) {
      out.println("<?xml version=\"1.0\"?>");
      out.println("<error>");
      out.println(e.getMessage());
      out.println("</error>");
    } finally {
      util.returnConnection(conn);
    }  
    
  }

  /** 
   * Handle documents passed to metacat that are encoded using the 
   * "multipart/form-data" mime type.  This is typically used for uploading
   * data files which may be binary and large.
   */
  private void handleMultipartForm(HttpServletRequest request,
                                   HttpServletResponse response) 
  {
    PrintWriter out = null;
    String action = null;

    // Parse the multipart form, and save the parameters in a Hashtable and
    // save the FileParts in a hashtable

    Hashtable params = new Hashtable();
    Hashtable fileList = new Hashtable();

    try {
      // MBJ: need to put filesize limit in Metacat config (metacat.properties)
      MultipartParser mp = new MultipartParser(request, 200*1024*1024); // 200MB
      Part part;
      while ((part = mp.readNextPart()) != null) {
        String name = part.getName();

        if (part.isParam()) {
          // it's a parameter part
          ParamPart paramPart = (ParamPart) part;
          String value = paramPart.getStringValue();
          params.put(name, value);
          if (name.equals("action")) {
            action = value;
          }
        } else if (part.isFile()) {
          // it's a file part
          FilePart filePart = (FilePart) part;
          fileList.put(name, filePart);

          // Stop once the first file part is found, otherwise going onto the
          // next part prevents access to the file contents.  So...for upload
          // to work, the datafile must be the last part
          break;
        }
      }
    } catch (IOException ioe) {
      try {
        out = response.getWriter();
      } catch (IOException ioe2) {
        System.err.println("Fatal Error: couldn't get response output stream.");
      }
      out.println("<?xml version=\"1.0\"?>");
      out.println("<error>");
      out.println("Error: problem reading multipart data.");
      out.println("</error>");
    }

    // Get the session information
    String username = null;
    String password = null;
    String[] groupnames = null;
    String sess_id = null;

    // be aware of session expiration on every request  
    HttpSession sess = request.getSession(true);
    if (sess.isNew()) {
      // session expired or has not been stored b/w user requests
      username = "public";
      sess.setAttribute("username", username);
    } else {
      username = (String)sess.getAttribute("username");
      password = (String)sess.getAttribute("password");
      groupnames = (String[])sess.getAttribute("groupnames");
      try {
        sess_id = (String)sess.getId();
      } catch(IllegalStateException ise) {
        System.out.println("error in  handleMultipartForm: this shouldn't " +
                           "happen: the session should be valid: " + 
                           ise.getMessage());
      }
    }  

    if ( action.equals("upload")) {
      if (username != null &&  !username.equals("public")) {
        handleUploadAction(request, response, params, fileList, 
                           username, groupnames);
      } else {
        try {
          out = response.getWriter();
        } catch (IOException ioe2) {
          System.err.println("Fatal Error: couldn't get response output stream.");
        }
        out.println("<?xml version=\"1.0\"?>");
        out.println("<error>");
        out.println("Permission denied for " + action);
        out.println("</error>");
      }
    } else {
      try {
        out = response.getWriter();
      } catch (IOException ioe2) {
        System.err.println("Fatal Error: couldn't get response output stream.");
      }
      out.println("<?xml version=\"1.0\"?>");
      out.println("<error>");
      out.println("Error: action not registered.  Please report this error.");
      out.println("</error>");
    }
  }

  /** 
   * Handle the upload action by saving the attached file to disk and 
   * registering it in the Metacat db
   */
  private void handleUploadAction(HttpServletRequest request,
                                  HttpServletResponse response, 
                                  Hashtable params, Hashtable fileList, 
                                  String username, String[] groupnames)
  {
    PrintWriter out = null;
    Connection conn = null;
    String action = null;
    String docid = null;

    response.setContentType("text/xml");
    try {
      out = response.getWriter();
    } catch (IOException ioe2) {
      System.err.println("Fatal Error: couldn't get response output stream.");
    }

    if (params.containsKey("docid")) {
      docid = (String)params.get("docid");
    }

    // Make sure we have a docid and datafile
    if (docid != null && fileList.containsKey("datafile")) {

      // Get a reference to the file part of the form
      FilePart filePart = (FilePart)fileList.get("datafile");
      String fileName = filePart.getFileName();
      MetaCatUtil.debugMessage("Uploading filename: " + fileName);

      // Check if the right file existed in the uploaded data
      if (fileName != null) {
  
        try {
          // register the file in the database (which generates an exception if
          // the docid is not acceptable or other untoward things happen
          DocumentImpl.registerDocument(fileName, "BIN", docid, username, 1);
    
          // Save the data file to disk using "docid" as the name
          dataDirectory.mkdirs();
          File newFile = new File(dataDirectory, docid);
          long size = filePart.writeTo(newFile);

          // set content type and other response header fields first
          out.println("<?xml version=\"1.0\"?>");
          out.println("<success>");
          out.println("<docid>" + docid + "</docid>"); 
          out.println("<size>" + size + "</size>"); 
          out.println("</success>");
    
        } catch (Exception e) {
          out.println("<?xml version=\"1.0\"?>");
          out.println("<error>");
          out.println(e.getMessage()); 
          out.println("</error>");
        }
      } else {
        // the field did not contain a file
        out.println("<?xml version=\"1.0\"?>");
        out.println("<error>");
        out.println("The uploaded data did not contain a valid file."); 
        out.println("</error>");
      }
    } else {
      // Error bcse docid missing or file missing
      out.println("<?xml version=\"1.0\"?>");
      out.println("<error>");
      out.println("The uploaded data did not contain a valid docid " +
                  "or valid file."); 
      out.println("</error>");
    }
  }
}
